// Contains class definitions
// Copyright (C) 1997 by John P. Weiss
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the Artistic License, included as the file
// "LICENSE" in the source code archive.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of the file "LICENSE", containing
// the License John Weiss originally placed this program under.
#include "AudioCD.h"
#include <stdlib.h>
#ifdef DEBUG_ON
#include <iostream.h>
#endif


// ***************
//  Here, we define the actual functions



//||||||||||||||||||||||||||||||||||||||||||||||||
//
//  Class AudioCD
//
//||||||||||||||||||||||||||||||||||||||||||||||||




//............................................................
//
// Inlined Methods
//
//............................................................


// References itemLengths via the playlist
//
inline int AudioCD::auPI_Len(int idx)
{
   return itemLengths[ToC->playlist[idx]];
}


// Handles pause and resume
//
inline void AudioCD::auTogglePause(void)
{
  player.PauseResume();

  if (player.getStatus() & cdsPaused)
	{
	emit newPlayState(psPaused);
	updateT->stop();
	}
  else
	{
	emit newPlayState(psPlaying);
	updateT->start(UPDATE_INTERVAL);
	}
}


// Resets all of the internal flags to an initial state
//
inline void AudioCD::auResetFlags(void)
{
  currentidx = nextidx = 0;
  repeat1 = repeat_all = shuffle = at_end = 
	do_repause = false;
}


// Wipes out the existing ToC and related lists.
//
inline void AudioCD::auEmptyToC(void)
{
   if (ToC != (Catalog *)0)
	 delete ToC;
   if (itemLengths != (int *)0)
	 delete [] itemLengths;

   ToC = (Catalog *)0;
   itemLengths = (int *)0;
   empty = true;
}


// Tries to reset the player device.  Closes the tray if it's open;
// re-inits the device if the player is empty.
//
inline void AudioCD::auResetPlayer(bool checkForEmpty)
{
  int pstat;

  pstat = player.getStatus();
  if (pstat & cdsEjected)
	{
	player.CloseTray();
	pstat = player.getStatus();
	}
  else if ( ((pstat & cdsEmpty)*(int)checkForEmpty) ||
		  (pstat & cdsError) )
	{
	player.InitDev();
	}

  auInitToC();
  if (!empty)
	{
	emit newItemNumber(ToC->len());
	emit newItemLabel( (char *)(ToC->name(0)) );
	emit newTime(proglen/CD_FRAMES);
	emit newPlayState(pmNormal | psOff);
	emit newProgLabel( (char *)(ToC->title()) );
	showArtist();
	}
}


// Emits a bunch of signals indicating that the player is empty
//
inline void AudioCD::auShowEmpty(void)
{
   emit newItemNumber(-1);
   emit newItemLabel("");
   emit newTime(-1);
   emit newPlayState(pmNormal| psOff);
   emit newProgLabel("No Disc");
   emit newProgArtist(0);
}


// Emits a new time, based on the current mode.
//
inline void AudioCD::auShowTime(int nutime)
{
  int out;

  if (show_total)
	{
	if (show_remain)
	   out = proglen - 
		 ToC->plTime(currentidx) - nutime;
	else
	   out = nutime + 
		 ToC->plTime(currentidx);
	}
  else
	{
	if (show_remain)
	   out = player.CurrentItemLength() - nutime;
	else
	   out = nutime;
	}
  emit newTime( out/CD_FRAMES );
}


// Emits a new index number and name [if the array is set up]
//
inline void AudioCD::auShowIndex(int nuidx)
{
  emit newItemNumber(ToC->playlist[nuidx]+1);
  emit newItemLabel(ToC->plName(nuidx));
}


// Creates, sets up, and connects the timers.
//
inline void AudioCD::auInitTimers(void)
{
  updateT = new QTimer( this );
  scanT = new QTimer( this );
  scanrateT = new QTimer( this );
  playnextT = new QTimer( this );
  
  connect( updateT, SIGNAL(timeout()), SLOT(auUpdateTime()) );
  connect( playnextT, SIGNAL(timeout()), SLOT(auPlayNext()) );
  connect( scanT, SIGNAL(timeout()), SLOT(auDoScan()) );
  connect( scanrateT, SIGNAL(timeout()), SLOT(auScanRate()) );
  return;
}


// Undoes a shuffle.
//
inline void AudioCD::auUnshuffle(void) 
{
  int i;
  for(i=0; i<ToC->pLen-1; i++)
	{
	ToC->playlist[i] = deshuffle_buf[i];
	}
  delete [] deshuffle_buf;
}


// Updates the program lengths
//
inline void AudioCD::auUpdateProgLen(void)
{
   int i;
   for(i=0, proglen=0; i<(ToC->pLen); i++)
	 proglen += itemLengths[i];
}



//.........................................................
//
// Public methods and Slots
//
//.........................................................



// Constructor for the AudioCD object.
//
AudioCD::AudioCD(void)
{
  // Initial values
  auResetFlags();
  do_clear_all = show_total = show_remain = false;
  itemLengths = (int *)0;
  empty = true;
}



// Starts up a new program
//
void AudioCD::newProgram(void)
{
   if(empty)
	 return;

   do_clear_all = false;
   StopPlay();
   auUpdateProgLen();
   StartPlayOrPause();
}



// Init the device for the first time.
//
void AudioCD::DeviceFirstInit( char *devname )
{
  int stat;

  auInitTimers();
  player.InitDev(devname);
  player.UpdateStatus();
  emit newVolume(player.Volume());

  // This method inits the ToC and updates everything but the volume
  // display.  The "false" tells the method not to check for an empty
  // CD [which would causes the drive to repeat open the tray].
  auResetPlayer(false);
  
  // Check the current device state and set the display.
  // Also starts up play if the device is currently playing or
  // paused. 
  stat = player.getStatus();
  if (stat & cdsPaused)
	{
	currentidx = player.CurrentItemIndex();
	player.lrqpSet(currentidx);
	auShowIndex(currentidx);
	emit newPlayState(pmNormal| psPaused);
	auShowTime(player.CurrentItemTime());
	updateT->start(UPDATE_INTERVAL);
#ifdef DEBUG_ON
	cerr << "%dbg%\tRestart:Paused - item #" 
		<< (ToC->playlist[currentidx]+1) << endl;
#endif
	}
  else if (stat & cdsPlaying)
	{
	// idx=0 <==> item #1
	currentidx = player.CurrentItemIndex();
	player.lrqpSet(currentidx);
	nextidx = currentidx + 1;
	if(nextidx > ToC->pLen-1)
	   at_end = true;
	auShowIndex(currentidx);
	emit newPlayState(pmNormal| psPlaying);
	auShowTime(player.CurrentItemTime());
	updateT->start(UPDATE_INTERVAL);
#ifdef DEBUG_ON
	cerr << "%dbg%\tRestart:Playing - item #" 
		<< (ToC->playlist[currentidx]+1) << endl;
#endif
	}
  else
	{
	currentidx = nextidx = 0;
	}
}



// The two slots which are the easiest to implement: the volume
// controls. 
//
void AudioCD::IncVol(void)
{
  player.IncVol();
  emit newVolume(player.Volume());
}
//
//
void AudioCD::DecVol(void)
{
  player.DecVol();
  emit newVolume(player.Volume());
}



// A high-end wrapper for the player-end eject/retract functions. 
//
void AudioCD::LoadOrEject(void)
{
  player.UpdateStatus();
  if (player.getStatus() & cdsEjected)
	{
	auResetPlayer();
	}
  else
	{
	// When we eject, stop the player, the timers, and re-init
	// EVERYTHING!
	player.Stop();
	updateT->stop();
	player.Eject();
	auResetFlags();
	auEmptyToC();
	auShowEmpty();
	}

}



// This stops all play.
//
void AudioCD::StopPlay(void)
{
  player.Stop();
  updateT->stop();

  if(empty)
	{
	auResetPlayer();
	return;
	}

  if (shuffle)
	auUnshuffle();
  
  if (do_clear_all)
	{
	do_clear_all = false;
	ToC->resetPlaylist();
	auUpdateProgLen();
	}
  else
	do_clear_all = true;

  auResetFlags();
  auShowIndex(0);
  emit newItemNumber(ToC->pLen);
  emit newTime( proglen/CD_FRAMES );
  emit newPlayState(psOff | pmNormal);
}



// This starts play at whereever the nextidx counter is, and in
// whatever mode we're currently in.  Since it's linked directly to
// the play/pause button, we need to test for that.
//
void AudioCD::StartPlayOrPause(void)
{
  do_clear_all = false;
  // For functions that access ToC[], always do this check.
  if(empty)
	{
	auResetPlayer();
	return;
	}

  if (player.getStatus() & cdsPlaying)
	{
	auTogglePause();
	return;
	}
	  
  if(shuffle)
	auShuffle();

  player.PlayItem(ToC->playlist[nextidx]);
  updateT->start(UPDATE_INTERVAL);
  currentidx = nextidx;
  nextidx++;
  if(nextidx > ToC->pLen-1)
	at_end = true;

  // idx=0 <==> item #1
  auShowIndex(currentidx);
  emit newPlayState(psPlaying);
}



// Moves to the next/previous track
//
void AudioCD::Next(void)
{
  do_clear_all = false;
  // For functions that access ToC[], always do this check.
  if(empty)
	{
	auResetPlayer();
	return;
	}

  nextidx = currentidx + 1;
  if(nextidx > ToC->pLen-1)
	{
	nextidx = ToC->pLen-1;
	return;
	}

  player.UpdateStatus();
  if (player.getStatus() & cdsPlaying)
	{
	if (player.getStatus() & cdsPaused)
	   do_repause = true;
	auPlayNext();
	}
  else
	{
	currentidx = nextidx;
	auShowIndex(currentidx);
	emit newTime(auPI_Len(currentidx)/CD_FRAMES);
	}
}
//
//
void AudioCD::Prev(void)
{
  double elt;

  do_clear_all = false;
  // For functions that access ToC[], always do this check.
  if(empty)
	{
	auResetPlayer();
	return;
	}

  nextidx = currentidx - 1;
  if(nextidx < 0)
	{
	nextidx = 0;
	}

  player.UpdateStatus();
  if (player.getStatus() == cdsPlaying) //It's not paused
	{
	at_end = false;
	elt = (double)(player.CurrentItemTime())/(double)CD_FRAMES;
	if(elt > 2.0) // 2 sec buffer.
	   {
	   nextidx = currentidx;
	   }
	auPlayNext();
	return;
	}

  if (player.getStatus() & cdsPaused)
	{
	at_end = false;
	do_repause = true;
	auPlayNext();
	return;
	}

  // Only gets done if we're not playing or paused - i.e. we're
  // stopped.
  currentidx = nextidx;
  auShowIndex(currentidx);
  emit newTime(auPI_Len(currentidx)/CD_FRAMES);
}



// Alters the time display mode
//
void AudioCD::newTimeMode(int mode)
{
  if(mode & tmAll)
	{
	show_total = true;
	}
  else
	{
	show_total = false;
	}
  
  if(mode & tmRemaining)
	{
	show_remain = true;
	}
  else
	{
	show_remain = false;
	}
}



// Alters the play mode.  If the new mode is a shuffle, it does the
// shuffle immediately.
//
void AudioCD::newPlayMode(int mode)
{
  switch(mode & PLAY_MODE_MASK)
	{
	 case pmNormal:
	    {
	    if (shuffle)
		  auUnshuffle();
	    
	    repeat1 = repeat_all = shuffle = false;
	    break;
	    }
	  //
	 case pmRepeatAll: 
	    {
	    if (shuffle)
		  auUnshuffle();
	    
	    repeat1 = shuffle = false;
	    repeat_all = true;
	    break;
	    }
	  //
	 case pmRepeat1:
	    {
	    repeat_all = shuffle = false;
	    repeat1 = true;
	    break;
	    }
	  //
	 case pmShuffle:
	    {
	    do_clear_all = false;
	    repeat1 = repeat_all = false;
	    shuffle = true;
	    auShuffle();
	    auPlayNext();
	    break;
	    }
	}
}



// Initializes a forward scan.
//
void AudioCD::StartScanFwd(void)
{
  int stat;

  do_clear_all = false;
  player.UpdateStatus();
  stat = player.getStatus();
  scan_not_playing = !(stat & cdsPlaying);
  if( empty || scan_not_playing )
	{
	return;
	}

  player.Stop();
  updateT->stop();

  if (stat & cdsPaused)
	{
	do_repause = true;
	}

#ifdef DEBUG_ON
  cerr << "%dbg%\tScan - Fwd\trepause=" << do_repause << endl;
#endif

  scantime = player.CurrentItemTime();
  scanstep = INITIAL_SCAN_STEP;
  scanning_forward = true;
  scanrateT->start(SCANRATE_INCR_INTERVAL);
  scanT->start(scanstep);
}



// Initializes a reverse scan.
//
void AudioCD::StartScanRev(void)
{
  int stat;

  do_clear_all = false;
  player.UpdateStatus();
  stat = player.getStatus();
  scan_not_playing = !(stat & cdsPlaying);
  if( empty || scan_not_playing )
	{
	return;
	}

  player.Stop();
  updateT->stop();

  if (stat & cdsPaused)
	{
	do_repause = true;
	}

#ifdef DEBUG_ON
  cerr << "%dbg%\tScan - Rev\trepause=" << do_repause << endl;
#endif

  scantime = player.CurrentItemTime();
  scanstep = INITIAL_SCAN_STEP;
  scanning_forward = false;
  scanrateT->start(SCANRATE_INCR_INTERVAL);
  scanT->start(scanstep);
}



// Stops a scan, resetting all ofhte variables that have changes,
// reinitializing play, and emitting new time/track values.
//
void AudioCD::StopScan(void)
{
  if(empty)
	{
	auResetPlayer();
	return;
	}

  // If we weren't playing, then there's nothing to stop when the
  // button-release signal comes.  This catches the case of an
  // attempted scan when the player wasn't playing.
  if( scan_not_playing )
 	{
	scan_not_playing = false;
 	return;
 	}

  // We'll only get this far if there's a scan to stop.
  scanT->stop();
  scanrateT->stop();
  updateT->start(UPDATE_INTERVAL);
#ifdef DEBUG_ON
  cerr << "%dbg%\tScan - Stop at: " << scantime/CD_FRAMES
	  << " sec" << endl;
#endif
  if(scantime >= auPI_Len(currentidx))
	{
#ifdef DEBUG_ON
	cerr << "%dbg%\tScan - Stop:toNext" << endl;
#endif
	scantime = 0;
	currentidx++;
	if(currentidx > ToC->pLen)
	   currentidx = 0;
	}
  player.PlayItemBlock(ToC->playlist[currentidx],scantime);
  auShowIndex(currentidx);

  // If it was paused, be sure to re-pause it.
  if (do_repause)
	{
	auTogglePause();
	do_repause = false;
	}
  return;
}



//............................................................
//
// Private Methods
//
//............................................................



// Performs the shuffling
//
void AudioCD::auShuffle(void) 
{
  int i,ridx1,ridx2, swapi;
  

  // This function can get called before the cd is playing, so we need
  // to start the update timer if it's inactive.
  if( !(updateT->isActive()) )
	updateT->start(UPDATE_INTERVAL);

  nextidx = 0;
  deshuffle_buf = new int[ToC->pLen];
  for(i=0; i < ToC->pLen-1; i++)
	deshuffle_buf[i] = ToC->playlist[i];

  for(i=0; i < 2*(ToC->pLen); i++)
	{
	ridx1 = (int)(drand48()*(ToC->pLen));
	ridx2 = (int)(drand48()*(ToC->pLen));

	swapi = ToC->playlist[ridx1];
	ToC->playlist[ridx1] = ToC->playlist[ridx2];
	ToC->playlist[ridx2] = swapi;
	}
}


// [Re]Inits the ToC from the device, loading a copy of the ToC.  This
// method assumes that the device is already open, has been
// initialized, and has read its internal ToC.
//
void AudioCD::auInitToC(void)
{
  int i, l;

  auEmptyToC();
  if (player.getStatus() & cdsEmpty)
	{
	auShowEmpty();
	empty = true;
	return;
	}

  l = player.ToC_Length();
  ToC = new Catalog(l);
  ToC->setTitle("No Title");
  empty = false;
  itemLengths = new int[l];
  
  ToC->setTime(l,player.ToC_Entry(l));
  for(i=0, proglen=0; i<l; i++)
	{
	ToC->setTime(i,player.ToC_Entry(i));
	itemLengths[i] = player.ItemLength(ToC->playlist[i]);
	proglen += itemLengths[i];
	}

  // Check to see if catalog is in the database.
  emit checkProgramCatalog();
  return;
}


// Resyncs the program with the player.  Only call this if the drive
// is currently running.
//
void AudioCD::auResyncProgram(void)
{
  int i, uidx;

  i = 0;
  uidx = -1;
  do
	{
	if(player.CurrentItemIndex() == ToC->playlist[i])
	   {
	   uidx = i;
	   i = ToC->pLen; //break the loop
	   }
	i++;
	}
  while(i < ToC->pLen);
  if(uidx == nextidx)
	{
	// We're playing the next track.  Are we repeating this track?
	if(repeat1)
	   {
	   // Yes: call auPlayNext
	   auPlayNext();
	   }
	else
	   {
	   // No, so update the counters.  We also update the lrqp
	   // counter in the device.  That way, if we're playing in
	   // sequence, the resync doesn't throw the player off two
	   // tracks from now.
	   currentidx = uidx;
	   player.lrqpSet(currentidx);
	   nextidx = currentidx + 1;
	   auShowIndex(currentidx);
	   if(nextidx > ToC->pLen-1)
		 {
		 at_end = true;
		 }
	   }
	}
  else if(uidx != currentidx)
	{
	// Okay, we're not playing the current track, and we're not
	// playing the next track.  So, we must've started playing the
	// next track on the disc, which isn't the next one on the
	// program.  No problem - just call auPlayNext().
	auPlayNext();
	}

  // Otherwise, do nothing.
}




//............................................................
//
// Private Slots
//
//............................................................



// Plays the next track in the program
//
void AudioCD::auPlayNext(void)
{
  // Check if we're at the end.
  if (at_end)
	{
	if(repeat_all)
	   {
#ifdef DEBUG_ON
	   cerr << "%dbg%\tauPlayNext - AtEnd:RepAll" << endl;
#endif
	   nextidx = 0;
	   at_end = false;
	   }
	else if (shuffle)
	   {
#ifdef DEBUG_ON
	   cerr << "%dbg%\tauPlayNext - AtEnd:Rep1" << endl;
#endif
	   auShuffle();
	   at_end = false;
	   }
	else	if(repeat1)
	   {
#ifdef DEBUG_ON
	   cerr << "%dbg%\tauPlayNext - AtEnd:Reshuffle" << endl;
#endif
	   nextidx = currentidx;
	   at_end = false;
	   }
	else
	   {
#ifdef DEBUG_ON
	   cerr << "%dbg%\tauPlayNext - AtEnd:Stop" << endl;
#endif
	   StopPlay();
	   return;
	   }
	}
  else if (repeat1)
	{
	nextidx = currentidx;
	}

#ifdef DEBUG_ON
  cerr << "%dbg%\tauPlayNext: old=" << ToC->playlist[currentidx]
	  << ", new=" << ToC->playlist[nextidx] << endl;
#endif


  currentidx = nextidx;
  nextidx = currentidx + 1;
  player.PlayItem(ToC->playlist[currentidx],true);
  auShowIndex(currentidx);

  // If it was paused, be sure to re-pause it.
  if (do_repause)
	{
	auTogglePause();
	do_repause = false;
	auShowTime(0);
	}

  // Set the end flag if necessary.
  if(nextidx > ToC->pLen-1)
	{
	at_end = true;
	}

  return;
}


// Get the current status and emit the signals to update the time.
// Also plays the next item if it's time.
//
void AudioCD::auUpdateTime(void)
{
  int ct,diff_ms;

  player.UpdateStatus();
  ct = player.CurrentItemTime();

  // First, check to see if we need to play the next track.  Have we
  // run into the next track?  Are we inside the playnext zone on
  // time?
  diff_ms = (auPI_Len(currentidx)-ct)*1000/CD_FRAMES; 
#ifdef DEBUG_ON
  if( diff_ms < 0 )
	{
	cerr << "%dbg%\tNegative diff=" << diff_ms 
		<< " msec ?!?!?!" << endl;
	}
#endif

  // ct can be negative if the drive thinks it's on the next track
  // but is still playing the last few frames from the previous
  // track.  In that event, we should resync the counters.
  // We also resync if the drive and the player don't agree on what
  // track their on.
  if ( (ct < 0) || (player.CurrentItemIndex() != ToC->playlist[currentidx]) )
	{
	auResyncProgram();
#ifdef DEBUG_ON
	cerr << "%dbg%\tResync: item #" << (ToC->playlist[currentidx]+1) << endl;
#endif
	}
  // Okay, so we *are* almost at the last track.  Set a timer to call
  // auPlayNext() when we're at the end.
  else if ( (diff_ms <= UPDATE_INTERVAL) && (ct > 0) )
	{
	if( !(playnextT->isActive()) )
	   {
	   if(diff_ms <= DT_PLAYNEXT_NOW)
		 {
		 // Oops - too close to start the timer. Call auPlayNext()
		 // now.
#ifdef DEBUG_ON
		 cerr << "%dbg%\tdiff=" << diff_ms << " msec" << endl;
#endif
		 auPlayNext();
		 }
	   else  // Set the timer.
		 {
		 diff_ms -= DT_PLAYNEXT_NOW;
		 playnextT->start(diff_ms,true);
#ifdef DEBUG_ON
		 cerr << "%dbg%\tplaynextT set for " 
			 << diff_ms << " msec" << endl;
#endif
		 }
	   }
	}


  // Phew!  We're done with that mess!  Now we can update the
  // display.
  auShowTime(ct);
}



// These next two slots handle scanning.  The last one is called by
// a separate timer and sets the rate of the timer that calls the
// first.
//
void AudioCD::auDoScan()
{
  if(scanning_forward)
	{
	scantime += CD_FRAMES;
	if(scantime > auPI_Len(currentidx))
	   scantime = auPI_Len(currentidx);
	}
  else
	{
	scantime -= CD_FRAMES;
	if(scantime < 0)
	   scantime = 0;
	}

  auShowTime(scantime);
  //  auShowIndex(currentidx);
}


// Change the scanning rate, dividing by 2 each time until we hit
// a lower limit.
//
void AudioCD::auScanRate()
{
  scanstep /= 2;
  if(scanstep < FASTEST_SCAN_STEP)
	scanstep = FASTEST_SCAN_STEP;

  scanT->changeInterval(scanstep);
#ifdef DEBUG_ON
  cerr << "%dbg%\tScanRateIncr: new rate=" << scanstep << endl;
#endif
}




//||||||||||||||||||||||||||||||||
//
// End
//
